@@ -23,6 +23,7 @@
 #include <stdio.h>
 #include <cstring>
 #include <string>
+#include <sys/stat.h>
 
 #include "../src/adplug.h"
 #include "../src/opl.h"
@@ -33,6 +34,9 @@
 #	define DIR_DELIM	"/"
 #endif
 
+#define TEST_SUBDIR  "testmus"
+#define REF_SUBDIR   "testref"
+
 /***** Local variables *****/
 
 // List of all filenames to test
@@ -49,16 +53,20 @@ static const char *filelist[] = {
   "BOOTUP.M",		// Ultima 6
   "CHILD1.XSM",		// eXtra Simple Music
   "DTM-TRK1.DTM",	// DeFy Adlib Tracker
-// Skip TwinTrack as it produces different results each time
-//  "fdance03.dmo",	// TwinTrack
+  "fdance03.dmo",	// TwinTrack
   "ice_thnk.sci",	// Sierra
   "inc.raw",		// RAW
   "crusader.raw",	// RAW (non-standard clock rate)
+  "GENORI.LD0",		// Loudness (older)
   "loudness.lds",	// Loudness
   "MARIO.A2M",		// AdLib Tracker 2
+  "AB_JULIA.A2T",	// Adlib Tracker 2 (tiny module v11)
+  "fank5.a2m",      // Adlib Tracker 2 (v11)
+  "fm-troni.a2m",   // Adlib Tracker 2 (v14)
   "mi2.laa",		// LucasArts
   "michaeld.cmf",	// Creative Music Format
   "2.CMF",			// Creative Music Format (with transpose effect)
+  "SNDTRACK.CMF",	// Creative Music Format (with note-off)
   "PLAYMUS1.SNG",	// SNGPlay
   "rat.xad",		// xad: rat
   "REVELAT.SNG",	// Faust Music Creator
@@ -71,6 +79,7 @@ static const char *filelist[] = {
   "TUBES.SAT",		// Surprise! Adlib Tracker
   "TU_BLESS.AMD",	// AMUSIC
   "VIB_VOL3.D00",	// EdLib Packed
+  "TheAlibi.d00",       // EdLib, reheadered
   "WONDERIN.WLF",	// Apogee
   "bmf1_1.bmf",		// BMF (without XAD header)
   "bmf1_2.xad",		// xad: BMF
@@ -83,8 +92,8 @@ static const char *filelist[] = {
   "blaster2.msc",	// AdLib MSCplay
   "RI051.RIX",		// Softstar RIX OPL Music
   "EOBSOUND.ADL",	// Westwood ADL v1
-  "DUNE19.ADL",		// Westwood ADL v2
-  "LOREINTR.ADL",	// Westwood ADL v3
+  "DUNE19.ADL",		// Westwood ADL v3
+  "LOREINTR.ADL",	// Westwood ADL v4
   "DEMO4.JBM",		// JBM Adlib Music
   "dro_v2.dro",         // DOSBox DRO v2.0
   "menu.got",       // God of Thunder Music (at 140 Hz)
@@ -104,14 +113,23 @@ static const char *filelist[] = {
   "MORNING.HSQ",	// HERAD SDB v1 (HSQ packed)
   "GORBI2.SQX",		// HERAD SDB v1 (SQX packed)
   "ARRAKIS.SDB",	// HERAD SDB v1
+  "SAVAGE.HSQ",		// HERAD SDB v1 (with instmode v2)
   "NEWSAN.HSQ",		// HERAD SDB v2 (HSQ packed)
   "NEWPAGA.HA2",	// HERAD SDB v2
   "WORMINTR.AGD",	// HERAD AGD
+  "well.adl",		// Coktel Vision ADL
+  "TEST16.MID",		// MIDI Type 0
+  "GRABBAG.MID",	// MIDI Type 1
+  "ACTION.PIS",		// Beni Tracker
+  "CHIP.MTR",		// Master Tracker v1
+  "AKMTEC.MTR",		// Master Tracker v2
+  "logical.plx",	// Palladix Sound System
   NULL
 };
 
 // String holding the relative path to the source directory
 static const char *testdir;
+static bool use_subdir = false;
 
 /***** Testopl *****/
 
@@ -142,7 +160,7 @@ public:
   {
     if(reg > 255 || val > 255 || reg < 0 || val < 0)
       std::cerr << "Warning: The player is writing data out of range! (reg = "
-		<< std::hex << reg << ", val = " << val << ")\n";
+    << std::hex << reg << ", val = " << val << ")" << std::dec << std::endl;
     if(!f) return;
     fprintf(f, "%x <- %x\n", reg, val);
   }
@@ -167,6 +185,18 @@ private:
 
 /***** Local functions *****/
 
+static bool dir_exists(std::string path)
+{
+  struct stat info;
+
+  int rc = stat(path.c_str(), &info);
+
+  if (rc != 0)
+    return false;
+
+  return (info.st_mode & S_IFDIR);
+}
+
 static bool diff(const std::string fn1, const std::string fn2)
   /*
    * Compares files 'fn1' and 'fn2' line by line and returns true if they are
@@ -209,10 +239,16 @@ static bool testplayer(const std::string filename)
   std::string	fn = std::string(testdir) + DIR_DELIM + filename;
 #ifdef __WATCOMC__
   std::string	testfn = tmpnam(NULL);
+#elif DJGPP
+  std::string	testfn = tmpnam(NULL);
 #else
   std::string	testfn = filename + ".test";
 #endif
   std::string	reffn = fn.substr(0, fn.find_last_of(".")) + ".ref";
+  if(use_subdir) {
+    fn.insert(fn.find_last_of(DIR_DELIM), DIR_DELIM TEST_SUBDIR);
+    reffn.insert(reffn.find_last_of(DIR_DELIM), DIR_DELIM REF_SUBDIR);
+  }
   Testopl	*opl = new Testopl(testfn);
   CPlayer	*p = CAdPlug::factory(fn, opl);
 
@@ -252,15 +288,23 @@ int main(int argc, char *argv[])
   testdir = getenv("testdir");
   if(!testdir) testdir = ".";
 
+  use_subdir = dir_exists(std::string(testdir) + DIR_DELIM TEST_SUBDIR);
+
+#ifdef DJGPP
+  // DJGPP/DosEMU has weird behaviour when parsing command line arguments (adds path and name of executable as seperate vars)
+  // Don't parse the rest by setting argument count to 1
+  std::cout << "Warning: Running test with DJGPP, ignoring command line arguments, executing all test files!" << std::endl;
+  argc = 1;
+#endif
   // Try all files one by one
   if(argc > 1) {
     for(i = 1; i < argc; i++)
       if(!testplayer(argv[i]))
-	retval = false;
-  } else
+        retval = false;
+  } else {
     for(i = 0; filelist[i] != NULL; i++)
       if(!testplayer(filelist[i]))
-	retval = false;
-
+        retval = false;
+  }
   return retval ? EXIT_SUCCESS : EXIT_FAILURE;
 }
